# Copyright 2020 Lorna Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import torch.nn as nn

from .utils import get_model_params
from .utils import load_pretrained_weights
from .utils import model_params


# Generative Adversarial Networks model architecture from
# the One weird trick...
# <https://arxiv.org/abs/1511.06434>`_ paper.
class Generator(nn.Module):
  r""" An Generator model. Most easily loaded with the .from_name
  or .from_pretrained methods

  Args:
    global_params (namedtuple): A set of GlobalParams shared between blocks

  Examples:
    >>> import torch
    >>> from dcgan_pytorch import Generator
    >>> from dcgan_pytorch import Discriminator
    >>> generator = Generator.from_pretrained("g-mnist")
    >>> discriminator = Discriminator.from_pretrained("g-mnist")
    >>> generator.eval()
    >>> noise = torch.randn(1, 100, 1, 1)
    >>> discriminator(generator(noise)).item()
  """

  def __init__(self, global_params=None):
    super(Generator, self).__init__()
    self.noise = global_params.noise
    self.channels = global_params.channels

    self.main = nn.Sequential(
      # inputs is Z, going into a convolution
      nn.ConvTranspose2d(self.noise, 64 * 8, 4, 1, 0, bias=False),
      nn.BatchNorm2d(64 * 8),
      nn.ReLU(True),
      # state size. (ngf*8) x 4 x 4
      nn.ConvTranspose2d(64 * 8, 64 * 4, 4, 2, 1, bias=False),
      nn.BatchNorm2d(64 * 4),
      nn.ReLU(True),
      # state size. (ngf*4) x 8 x 8
      nn.ConvTranspose2d(64 * 4, 64 * 2, 4, 2, 1, bias=False),
      nn.BatchNorm2d(64 * 2),
      nn.ReLU(True),
      # state size. (ngf*2) x 16 x 16
      nn.ConvTranspose2d(64 * 2, 64, 4, 2, 1, bias=False),
      nn.BatchNorm2d(64),
      nn.ReLU(True),
      # state size. (ngf) x 32 x 32
      nn.ConvTranspose2d(64, self.channels, 4, 2, 1, bias=False),
      nn.Tanh()
      # state size. (nc) x 64 x 64
    )

  def forward(self, inputs):
    outputs = self.main(inputs)
    return outputs

  @classmethod
  def from_name(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    global_params = get_model_params(model_name)
    return cls(global_params)

  @classmethod
  def from_pretrained(cls, model_name):
    model = cls.from_name(model_name, )
    load_pretrained_weights(model, model_name)
    return model

  @classmethod
  def get_image_size(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    _, res = model_params(model_name)
    return res

  @classmethod
  def _check_model_name_is_valid(cls, model_name):
    """ Validates model name. """
    valid_list = ["mnist", "fmnist", "cifar", "imagenet"]
    valid_models = ["g-" + str(i) for i in valid_list]
    if model_name not in valid_models:
      raise ValueError("model_name should be one of: " + ", ".join(valid_models))


class Discriminator(nn.Module):
  r""" An Discriminator model. Most easily loaded with the .from_name or .from_pretrained methods

  Args:
    global_params (namedtuple): A set of GlobalParams shared between blocks

  Examples:
    >>> import torch
    >>> from dcgan_pytorch import Discriminator
    >>> discriminator = Discriminator.from_pretrained("d-mnist")
    >>> discriminator.eval()
    >>> noise = torch.randn(1, 784, 1, 1)
    >>> discriminator(noise).item()
  """

  def __init__(self, global_params=None):
    super(Discriminator, self).__init__()
    self.channels = global_params.channels
    self.image_size = global_params.image_size
    self.negative_slope = global_params.negative_slope

    self.main = nn.Sequential(
      # inputs is (nc) x 64 x 64
      nn.Conv2d(self.channels, 64, 4, 2, 1, bias=False),
      nn.LeakyReLU(self.negative_slope, inplace=True),
      # state size. (ndf) x 32 x 32
      nn.Conv2d(64, 64 * 2, 4, 2, 1, bias=False),
      nn.BatchNorm2d(64 * 2),
      nn.LeakyReLU(self.negative_slope, inplace=True),
      # state size. (ndf*2) x 16 x 16
      nn.Conv2d(64 * 2, 64 * 4, 4, 2, 1, bias=False),
      nn.BatchNorm2d(64 * 4),
      nn.LeakyReLU(self.negative_slope, inplace=True),
      # state size. (ndf*4) x 8 x 8
      nn.Conv2d(64 * 4, 64 * 8, 4, 2, 1, bias=False),
      nn.BatchNorm2d(64 * 8),
      nn.LeakyReLU(self.negative_slope, inplace=True),
      # state size. (ndf*8) x 4 x 4
      nn.Conv2d(64 * 8, 1, 4, 1, 0, bias=False),
      nn.Sigmoid()
    )

  def forward(self, inputs):
    outputs = self.main(inputs)
    outputs = torch.flatten(outputs)
    return outputs

  @classmethod
  def from_name(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    global_params = get_model_params(model_name)
    return cls(global_params)

  @classmethod
  def from_pretrained(cls, model_name):
    model = cls.from_name(model_name, )
    load_pretrained_weights(model, model_name)
    return model

  @classmethod
  def get_image_size(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    _, res = model_params(model_name)
    return res

  @classmethod
  def _check_model_name_is_valid(cls, model_name):
    """ Validates model name. """
    valid_list = ["mnist", "fmnist", "cifar", "imagenet"]
    valid_models = ["d-" + str(i) for i in valid_list]
    if model_name not in valid_models:
      raise ValueError("model_name should be one of: " + ", ".join(valid_models))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 Lorna Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'dcgan_pytorch'
DESCRIPTION = 'PyTorch implements a simple GAN neural network structure.'
URL = 'https://github.com/lornatang/DCGAN-PyTorch'
EMAIL = 'liuchangyu1111@gmail.com'
AUTHOR = 'Liu Changyu'
REQUIRES_PYTHON = '>=3.5.0'
VERSION = '0.1.4'

# What packages are required for this module to be executed?
REQUIRED = [
  'torch',
]

# What packages are optional?
EXTRAS = {
  # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier
# for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
  with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()
except FileNotFoundError:
  long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
  project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
  with open(os.path.join(here, project_slug, '__version__.py')) as f:
    exec(f.read(), about)
else:
  about['__version__'] = VERSION


class UploadCommand(Command):
  """Support setup.py upload."""

  description = 'Build and publish the package.'
  user_options = []

  @staticmethod
  def status(s):
    """Prints things in bold."""
    print('\033[1m{0}\033[0m'.format(s))

  def initialize_options(self):
    pass

  def finalize_options(self):
    pass

  def run(self):
    try:
      self.status('Removing previous builds…')
      rmtree(os.path.join(here, 'dist'))
    except OSError:
      pass

    self.status('Building Source and Wheel (universal) distribution…')
    os.system(
      '{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

    self.status('Uploading the package to PyPI via Twine…')
    os.system('twine upload dist/*')

    self.status('Pushing git tags…')
    os.system('git tag v{0}'.format(about['__version__']))
    os.system('git push --tags')

    sys.exit()


# Where the magic happens:
setup(
  name=NAME,
  version=about['__version__'],
  description=DESCRIPTION,
  long_description=long_description,
  long_description_content_type='text/markdown',
  author=AUTHOR,
  author_email=EMAIL,
  python_requires=REQUIRES_PYTHON,
  url=URL,
  packages=find_packages(
    exclude=[
      "tests",
      "*.tests",
      "*.tests.*",
      "tests.*"]),
  # py_modules=['model'], # If your package is a single module, use this
  # instead of 'packages'
  install_requires=REQUIRED,
  extras_require=EXTRAS,
  include_package_data=True,
  license='Apache',
  classifiers=[
    # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
  ],
  # $ setup.py publish support.
  cmdclass={
    'upload': UploadCommand,
  },
)



# Copyright 2020 Lorna Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import collections
import ssl

import torch.utils.model_zoo as model_zoo

# Parameters for the entire model (stem, all blocks, and head)
GlobalParams = collections.namedtuple("GlobalParams", [
  "noise", "channels", "image_size",
  "negative_slope"
])

# Change namedtuple defaults
GlobalParams.__new__.__defaults__ = (None,) * len(GlobalParams._fields)


# custom weights initialization called on netG and netD
def weights_init(m):
  classname = m.__class__.__name__
  if classname.find('Conv') != -1:
    m.weight.data.normal_(0.0, 0.02)
  elif classname.find('BatchNorm') != -1:
    m.weight.data.normal_(1.0, 0.02)
    m.bias.data.fill_(0)


########################################################################
############## HELPERS FUNCTIONS FOR LOADING MODEL PARAMS ##############
########################################################################


def model_params(model_name):
  r""" Map Generator and Discriminator model name to parameter coefficients.

    Args:
        model_name (string): The name of the model corresponding to the dataset.
    Return:
        A binary tuple value.
    """
  params_dict = {
    # Coefficients: channels, image_size, layer
    "g-mnist": (1, 64),
    "g-fmnist": (1, 64),
    "g-cifar": (3, 64),
    "g-imagenet": (3, 64),
    "d-mnist": (1, 64),
    "d-fmnist": (1, 64),
    "d-cifar": (3, 64),
    "d-imagenet": (3, 64),
  }
  return params_dict[model_name]


def model(channels=None, image_size=None):
  r""" Gets the parameters of the model

    Args:
        channels (int): size of each input image channels.
        image_size (int): size of each input image size.

    Return:
        A set of GlobalParams shared between blocks.
    """
  global_params = GlobalParams(
    noise=100,
    negative_slope=0.2,
    channels=channels,
    image_size=image_size,
  )

  return global_params


def get_model_params(model_name):
  """ Get the block args and global params for a given model

    Args:
        model_name (string): The name of the model corresponding to the dataset.
    Return:
        A set of GlobalParams shared between blocks.
    """
  if model_name.startswith("g") or model_name.startswith("d"):
    c, s = model_params(model_name)
    global_params = model(channels=c, image_size=s)
  else:
    raise NotImplementedError(f"model name is not pre-defined: {model_name}.")
  return global_params


url_maps = {
  "d-cifar": "https://github.com/changyu98/models/raw/master/pytorch/dcgan/d-cifar-d90b69d8.pth",
  "d-fmnist": "https://github.com/changyu98/models/raw/master/pytorch/dcgan/d-fmnist-46a7c6bc.pth",
  "d-mnist": "https://github.com/changyu98/models/raw/master/pytorch/dcgan/d-mnist-71297c87.pth",
  "g-cifar": "https://github.com/changyu98/models/raw/master/pytorch/dcgan/g-cifar-84f37fe1.pth",
  "g-fmnist": "https://github.com/changyu98/models/raw/master/pytorch/dcgan/g-fmnist-6a6da537.pth",
  "g-mnist": "https://github.com/changyu98/models/raw/master/pytorch/dcgan/g-mnist-b55a0fb7.pth",
}


def load_pretrained_weights(model_arch, model_name):
  """ Loads pretrained weights, and downloads if loading for the first time. """

  try:
    _create_unverified_https_context = ssl._create_unverified_context
  except AttributeError:
    # Legacy Python that doesn't verify HTTPS certificates by default
    pass
  else:
    # Handle target environment that doesn't support HTTPS verification
    ssl._create_default_https_context = _create_unverified_https_context

  state_dict = model_zoo.load_url(url_maps[model_name])
  model_arch.load_state_dict(state_dict)
  print(f"Loaded model pretrained weights for `{model_name}`.")
